"""A generic OIDC client that can use OOB or not."""
from functools import partial
from http.server import BaseHTTPRequestHandler, HTTPServer
import json
import os
import re
import threading

from authlib.integrations.base_client.errors import OAuthError
from authlib.integrations.requests_client import OAuth2Session
from authlib.oidc.discovery.well_known import get_well_known_url
from requests_kerberos import HTTPKerberosAuth, OPTIONAL
import click
import requests


PORT = 45678  # Hopefully nothing else uses this on the host...

RESULT_SUCCESS = """
<html>
    <head>
        <title>Authentication successful!</title>
        <style>
            p {
                margin-top: 2em;
                font-size: 130%;
                text-align: center;
            }
        </style>
    </head>
    <body>
        <p>
            <strong>Authentication successful!</strong>
            You can now close this browser window and go back to bodhi's command-line.
        </p>
    </body>
</html>
"""


class OIDCClientError(Exception):
    """Raised when there's an error in the OIDCClient class."""


class OIDCClient:
    """A client for OpenID Connect authentication."""

    def __init__(
        self, client_id, scope, id_provider, storage,
    ):
        """Initialize OIDCClient.

        Args:
            client_id (str): The OIDC client ID.
            scope (str): The OIDC scopes that will be asked.
            id_provider (str): The URL to the OIDC provider.
            storage (JSONStorage): An instance of JSONStorage to store the tickets.
        """
        self.client_id = client_id
        self.scope = scope
        self.id_provider = id_provider
        self.storage = storage
        self._tokens = None
        self._username = None
        self.metadata = {}
        self._build_client(client_id, scope, id_provider)

    def _get_provider_metadata(self, id_provider):
        metadata_endpoint = get_well_known_url(id_provider, external=True)
        response = requests.get(metadata_endpoint)
        if not response.ok:
            raise OIDCClientError(response.text)
        self.metadata = response.json()

    def _build_client(self, client_id, scope, id_provider):
        self._get_provider_metadata(id_provider)
        self.client = OAuth2Session(
            client_id,
            scope=scope,
            token_endpoint_auth_method="none",
            redirect_uri=self.redirect_uri,
            update_token=self._update_token,
            token_endpoint=self.metadata["token_endpoint"],
            token=self.tokens,
        )

    @property
    def _use_oob(self):
        return "oob" in self.metadata.get("response_modes_supported", [])

    @property
    def redirect_uri(self):
        """Return the OIDC redirect URI.

        The value will depend on the server's ability to do OOB authentication.

        Returns:
            str: The OIDC redirect URI.
        """
        if self._use_oob:
            return "urn:ietf:wg:oauth:2.0:oob"
        else:
            return f"http://localhost:{PORT}/auth"

    @property
    def username(self):
        """Return the authenticated user name.

        This will trigger authentication.

        Returns:
            str or None: The authenticated username or ``None`` if we couldn't authenticate.
        """
        if self._username is None:
            self.ensure_auth()
            response = self.client.get(self.metadata["userinfo_endpoint"])
            if response.ok:
                self._username = response.json()["nickname"]
        return self._username

    def login(self, use_kerberos=False):
        """Login to the OIDC provider.

        If authentication fails, it will be retried.

        Args:
            use_kerberos (bool): Use Kerberos for authentication.

        Raises:
            click.ClickException: When authentication was cancelled.
        """
        authorization_endpoint = self.metadata["authorization_endpoint"]
        uri, state_ = self.client.create_authorization_url(authorization_endpoint)
        # 1. use_kerberos is True and Kerberos succeeds -> print success
        # 2. use_kerberos is True and Kerberos fails -> browser login follows
        # 3. use_kerberos is False -> browser login only
        if use_kerberos:
            try:
                self.login_with_kerberos(uri)
            except (OIDCClientError, OAuthError) as e:
                click.secho(
                    f"Kerberos authentication failed ({e}). "
                    f"Proceeding with browser-based authentication.",
                    fg="red"
                )
        if not self.is_logged_in:
            self.login_with_browser(uri)
        click.secho("Login successful!", fg="green")

    def login_with_browser(self, uri):
        """Login to the OIDC provider using the browser.

        If authentication failed, it will be retried.

        Args:
            uri (str): Authentication URL as obtained from ``create_authorization_url()``.

        Raises:
            click.ClickException: Login has been cancelled.
        """
        click.secho("Authenticating... Please open your browser to:", fg="yellow")
        click.echo(uri)
        if self._use_oob:
            while True:
                try:
                    value = click.prompt(
                        click.style(
                            "Paste here the code that you got after logging in", fg="yellow"
                        ),
                        hide_input=True,
                    )
                except click.exceptions.Abort:
                    # Raise a generic exception for friendliness to outside scripts
                    # that don't want to import click
                    raise SystemExit("Cancelled.")
                try:
                    self.auth_callback(f"?{value}")
                except OAuthError as e:
                    click.secho(f"Login failed!: {e}. Please try again.", fg="red")
                if self.tokens:
                    break
        else:
            self._run_http_server()

    def login_with_kerberos(self, uri):
        """Login to the OIDC provider using Kerberos.

        Raises:
            OIDCClientError: if there is a problem during the auth process.
        """
        response = requests.get(
            uri,
            auth=HTTPKerberosAuth(
                # REQUIRED is not working with id.fedoraproject.org
                mutual_authentication=OPTIONAL,
            ),
        )
        try:
            response.raise_for_status()
        except requests.RequestException as e:
            raise OIDCClientError(
                f"There was an issue while performing Kerberos authentication: {e}")
        try:
            value = re.findall(
                r"<title>\s*(code=[\w\-_=;&]+)\s*</title>", response.text
            )[0]
        except IndexError:
            raise OIDCClientError(
                f'Unable to locate OIDC code in the response from "{uri}".'
            )
        self.auth_callback(f"?{value}")

    def _run_http_server(self):
        httpd = HTTPServer(
            ("localhost", PORT), partial(RequestHandler, callback=self.auth_callback)
        )
        server_thread = threading.Thread(target=httpd.serve_forever)
        # Exit the server thread when the main thread terminates
        server_thread.daemon = True
        server_thread.start()
        try:
            server_thread.join()
        except KeyboardInterrupt:
            stop_thread = threading.Thread(target=httpd.shutdown, daemon=True)
            stop_thread.start()
            stop_thread.join()
            raise click.ClickException("Cancelled.")
        finally:
            httpd.server_close()

    def auth_callback(self, response):
        """Handle OIDC callback (post-login).

        Args:
            response (str): The authorization response that the OIDC provider sent.
        """
        self.tokens = self.client.fetch_token(
            self.metadata["token_endpoint"],
            authorization_response=response,
            redirect_uri=self.redirect_uri,
        )

    @property
    def is_logged_in(self):
        """Check whether the client is logged in with the provider."""
        return bool(self.tokens)

    def ensure_auth(self, use_kerberos=False):
        """Make sure the client is authenticated."""
        if not self.is_logged_in:
            self.login(use_kerberos=use_kerberos)

    @property
    def tokens(self):
        """Return the authentication tokens, or None if we don't have any yet.

        Returns:
            dict: The authentication tokens.
        """
        if self._tokens is None:
            self._tokens = self.storage.load("tokens", {}).get(self.id_provider)
        return self._tokens

    @tokens.setter
    def tokens(self, value):
        self._tokens = value
        stored_tokens = self.storage.load("tokens", {})
        stored_tokens[self.id_provider] = value
        self.storage.save("tokens", stored_tokens)

    def _update_token(self, token, refresh_token=None, access_token=None):
        self.tokens = token

    def request(self, *args, **kwargs):
        """Make an authenticated request.

        The request will have a Bearer authentication token using the OIDC access_token.

        Raises:
            OIDCClientError: When the connection fails

        Returns:
            Response: The request's response.
        """
        self.ensure_auth()
        try:
            return self.client.request(*args, **kwargs)
        except requests.exceptions.ConnectionError as e:
            raise OIDCClientError(str(e))
        except OAuthError:
            # Auth failed, clear it and retry
            self.clear_auth()
            return self.request(*args, **kwargs)

    def has_cookie(self, name, domain=None):
        """Return whether the OIDC client has a cookie of the provided name.

        Args:
            name (str): The name of the cookie.
            domain (str or None): The domain of the cookie. Defaults to None for any domain.

        Returns:
            bool: True if the HTTP client has this cookie, False otherwise.
        """
        if domain is not None and "." not in domain:
            domain += ".local"
        return self.client.cookies.get(name, domain=domain) is not None

    def clear_auth(self):
        """Clear the authentication tokens and cache."""
        self.tokens = {}
        self._username = None
        self.client.token_auth.set_token({})
        self.client.cookies.clear()


class RequestHandler(BaseHTTPRequestHandler):
    """A request handler for the embedded HTTP server."""

    def __init__(self, *args, **kwargs):
        """Initialize the request handler."""
        self.callback = kwargs.pop("callback")
        super().__init__(*args, **kwargs)

    def do_GET(self):
        """Handle GET requests."""
        try:
            self.callback(self.path)
        except Exception as e:
            click.echo(e, err=True)
            self.send_error(500, str(e))
        else:
            self.send_response(200)
            self.send_header("Content-type", "text/html")
            self.end_headers()
            self.wfile.write(RESULT_SUCCESS.encode("utf-8"))
        threading.Thread(target=self.server.shutdown, daemon=True).start()


class JSONStorage:
    """Store dictionaries as JSON in a regular file."""

    def __init__(self, path):
        """Initialize the JSONStorage object.

        Args:
            path (str): The path to the JSON file.
        """
        self.path = path

    def load_all(self):
        """Load all the data from the file.

        Returns:
            dict: The stored data.
        """
        if not os.path.exists(self.path):
            return {}
        with open(self.path) as f:
            return json.load(f)

    def load(self, key, default=None):
        """Load a specific key from the storage.

        Args:
            key (str): The key to load
            default (any, optional): What to return if the key is not found in the storage.
                Defaults to None.

        Returns:
            any: The stored value for the specified key.
        """
        return self.load_all().get(key, default)

    def save(self, key, value):
        """Save a value in the store.

        Args:
            key (str): The key to store the value as.
            value (any): The value to store. It must be JSON-serializable.
        """
        data = self.load_all()
        data[key] = value
        parent_dir = os.path.dirname(self.path)
        if not os.path.exists(parent_dir):
            os.makedirs(parent_dir)
        with open(self.path, "w") as f:
            json.dump(data, f)



# Copyright © 2016-2019 Red Hat, Inc. and others.
#
# This file is part of Bodhi.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""This module contains test data for bodhi.client."""

from munch import Munch


EXAMPLE_COMMENT_MUNCH = Munch({
    'comment': Munch({
        'bug_feedback': [], 'user_id': 2897, 'author': 'bowlofeggs',
        'timestamp': '2017-02-28 18:08:13', 'testcase_feedback': [], 'karma_critpath': 0,
        'update': Munch({
            'date_testing': '2017-02-13 22:56:06', 'pushed': True,
            'require_testcases': False, 'date_stable': None, 'critpath': False,
            'date_approved': None, 'stable_karma': 3, 'date_pushed': '2017-02-13 22:56:06',
            'requirements': '', 'severity': 'low', 'autokarma': True, 'autotime': True,
            'title': 'nodejs-grunt-wrap-0.3.0-2.fc25', 'suggest': 'unspecified',
            'require_bugs': False, 'date_locked': None, 'type': 'newpackage',
            'close_bugs': True, 'status': 'testing', 'meets_testing_requirements': True,
            'date_submitted': '2017-02-13 17:38:43', 'unstable_karma': -3,
            'user': Munch({
                'openid': 'bowlofeggs.id.fedoraproject.org', 'name': 'bowlofeggs',
                'id': 2897,
                'avatar': 'AVATAR_URL',
                'groups': [Munch({'name': 'packager'})],
                'email': 'email@example.com'}),
            'locked': False,
            'builds': [
                Munch({'epoch': 0, 'nvr': 'nodejs-grunt-wrap-0.3.0-2.fc25', 'signed': True})],
            'date_modified': None, 'url': 'http://localhost:6543/updates/FEDORA-2017-c95b33872d',
            'notes': 'New package.', 'request': None,
            'bugs': [Munch({
                'bug_id': 1420605, 'security': False, 'feedback': [], 'parent': False,
                'title': ('Review Request: nodejs-grunt-wrap - A Grunt plugin for wrapping '
                          'project text files')})],
            'alias': 'FEDORA-2017-c95b33872d', 'karma': 0,
            'release': Munch({
                'dist_tag': 'f25', 'name': 'F25', 'testing_tag': 'f25-updates-testing',
                'pending_stable_tag': 'f25-updates-pending',
                'pending_signing_tag': 'f25-signing-pending', 'long_name': 'Fedora 25',
                'state': 'current', 'version': '25', 'override_tag': 'f25-override',
                'branch': 'f25', 'id_prefix': 'FEDORA',
                'pending_testing_tag': 'f25-updates-testing-pending',
                'stable_tag': 'f25-updates', 'candidate_tag': 'f25-updates-candidate',
                'package_manager': 'unspecified', 'testing_repository': None, 'eol': None})}),
        'update_id': 79733, 'karma': 0, 'text': 'i found $10000',
        'update_title': 'nodejs-grunt-wrap-0.3.0-2.fc25', 'id': 562626,
        'user': Munch({
            'openid': 'bowlofeggs.id.fedoraproject.org', 'name': 'bowlofeggs',
            'id': 2897, 'avatar': 'AVATAR_URL',
            'groups': [Munch({'name': 'packager'})], 'email': 'email@example.com'})}),
    'caveats': []})

# EXAMPLE_COMMENT_MUNCH is expected to generate this output in update_str
EXPECTED_COMMENT_OUTPUT = """The following comment was added to FEDORA-2017-c95b33872d
i found $10000
"""


EXPECTED_COMPOSE_OUTPUT = """\
================================================================================
     *EPEL-7-stable  :   2 updates (requested)
================================================================================
Content Type: rpm
     Started: 2018-03-15 17:25:22
     Updated: 2018-03-15 17:25:22

Updates:

        FEDORA-EPEL-2018-50566f0a39: uwsgi-2.0.16-1.el7
        FEDORA-EPEL-2018-328e2b8c27: qtpass-1.2.1-3.el7

"""


EXAMPLE_COMPOSE_MUNCH = Munch({
    'compose': Munch({
        'release_id': 8, 'content_type': 'rpm',
        'update_summary': [
            Munch({'alias': 'FEDORA-EPEL-2018-50566f0a39', 'title': 'uwsgi-2.0.16-1.el7'}),
            Munch({'alias': 'FEDORA-EPEL-2018-328e2b8c27', 'title': 'qtpass-1.2.1-3.el7'})],
        'error_message': None, 'request': 'stable', 'state': 'requested',
        'state_date': '2018-03-15 17:25:22', 'checkpoints': '{}',
        'release': Munch({
            'dist_tag': 'epel7', 'name': 'EPEL-7', 'testing_tag': 'epel7-testing',
            'pending_stable_tag': 'epel7-pending',
            'pending_signing_tag': 'epel7-signing-pending', 'long_name': 'Fedora EPEL 7',
            'state': 'current', 'version': '7', 'override_tag': 'epel7-override',
            'branch': 'epel7', 'id_prefix': 'FEDORA-EPEL',
            'pending_testing_tag': 'epel7-testing-pending', 'stable_tag': 'epel7',
            'candidate_tag': 'epel7-testing-candidate',
            'package_manager': 'unspecified', 'testing_repository': None,
            'released_on': None, 'eol': None, 'setting_status': None}),
        'date_created': '2018-03-15 17:25:22', 'security': True})})


EXAMPLE_COMPOSES_MUNCH = Munch({
    'composes': [
        Munch({
            'release_id': 8, 'content_type': 'rpm',
            'update_summary': [
                Munch({'alias': 'FEDORA-EPEL-2018-50566f0a39', 'title': 'uwsgi-2.0.16-1.el7'}),
                Munch({'alias': 'FEDORA-EPEL-2018-328e2b8c27', 'title': 'qtpass-1.2.1-3.el7'})],
            'error_message': None, 'request': 'stable', 'state': 'requested',
            'state_date': '2018-03-15 17:25:22', 'checkpoints': '{}',
            'release': Munch({
                'dist_tag': 'epel7', 'name': 'EPEL-7', 'testing_tag': 'epel7-testing',
                'pending_stable_tag': 'epel7-pending',
                'pending_signing_tag': 'epel7-signing-pending', 'long_name': 'Fedora EPEL 7',
                'state': 'current', 'version': '7', 'override_tag': 'epel7-override',
                'branch': 'epel7', 'id_prefix': 'FEDORA-EPEL',
                'pending_testing_tag': 'epel7-testing-pending', 'stable_tag': 'epel7',
                'candidate_tag': 'epel7-testing-candidate',
                'package_manager': 'unspecified', 'testing_repository': None,
                'released_on': None, 'eol': None, 'setting_status': None}),
            'date_created': '2018-03-15 17:25:22', 'security': True}),
        Munch({
            'release_id': 8, 'content_type': 'rpm',
            'update_summary': [
                Munch({'alias': 'FEDORA-EPEL-2018-32f78e466c',
                       'title': 'libmodulemd-1.1.0-1.el7'})],
            'error_message': None, 'request': 'testing', 'state': 'requested',
            'state_date': '2018-03-15 17:25:22', 'checkpoints': '{}',
            'release': Munch({
                'dist_tag': 'epel7', 'name': 'EPEL-7', 'testing_tag': 'epel7-testing',
                'pending_stable_tag': 'epel7-pending',
                'pending_signing_tag': 'epel7-signing-pending', 'long_name': 'Fedora EPEL 7',
                'state': 'current', 'version': '7', 'override_tag': 'epel7-override',
                'branch': 'epel7', 'id_prefix': 'FEDORA-EPEL',
                'pending_testing_tag': 'epel7-testing-pending', 'stable_tag': 'epel7',
                'candidate_tag': 'epel7-testing-candidate',
                'package_manager': 'unspecified', 'testing_repository': None,
                'released_on': None, 'eol': None, 'setting_status': None}),
            'date_created': '2018-03-15 17:25:22', 'security': False})]})


EXAMPLE_OVERRIDE_MUNCH = Munch({
    'build_id': 108570, 'submission_date': '2017-02-28 23:05:32', 'caveats': [],
    'nvr': 'js-tag-it-2.0-1.fc25', 'expiration_date': '2017-03-07 23:05:31',
    'notes': 'No explanation given...', 'submitter_id': 2897,
    'build': Munch(
        {'epoch': 0, 'nvr': 'js-tag-it-2.0-1.fc25', 'signed': True, 'release_id': 15}),
    'expired_date': None, 'submitter': Munch({
        'openid': None, 'name': 'bowlofeggs', 'id': 2897, 'avatar': None,
        'groups': [Munch({'name': 'packager'})], 'email': 'email@example.com'})})

EXAMPLE_OVERRIDE_MUNCH_CAVEATS = Munch({
    'build_id': 108570, 'submission_date': '2017-02-28 23:05:32',
    'caveats': [Munch({'description': 'this is a caveat'})],
    'nvr': 'js-tag-it-2.0-1.fc25', 'expiration_date': '2017-03-07 23:05:31',
    'notes': 'No explanation given...', 'submitter_id': 2897,
    'build': Munch({'epoch': 0, 'nvr': 'js-tag-it-2.0-1.fc25', 'signed': True}),
    'expired_date': None, 'submitter': Munch({
        'openid': None, 'name': 'bowlofeggs', 'id': 2897, 'avatar': None,
        'groups': [Munch({'name': 'packager'})], 'email': 'email@example.com'})})

EXAMPLE_EXPIRED_OVERRIDE_MUNCH = Munch({
    'build_id': 108570, 'submission_date': '2017-02-28 23:05:32', 'caveats': [],
    'nvr': 'js-tag-it-2.0-1.fc25', 'expiration_date': '2017-03-07 23:05:31',
    'notes': 'This is an expired override', 'submitter_id': 2897,
    'build': Munch({'epoch': 0, 'nvr': 'js-tag-it-2.0-1.fc25', 'signed': True}),
    'expired_date': '2017-03-07 23:05:31', 'submitter': Munch({
        'openid': None, 'name': 'bowlofeggs', 'id': 2897, 'avatar': None,
        'groups': [Munch({'name': 'packager'})], 'email': 'email@example.com'})})

EXAMPLE_QUERY_MUNCH = Munch({
    'chrome': True,
    'display_request': True,
    'display_user': True,
    'package': None,
    'page': 1,
    'pages': 1,
    'rows_per_page': 20,
    'total': 1,
    'updates': [{
        'alias': 'FEDORA-2017-c95b33872d',
        'autokarma': True,
        'autotime': True,
        'content_type': 'rpm',
        'bugs': [{
            'bug_id': 1420605,
            'feedback': [],
            'parent': False,
            'security': False,
            'title': ('Review Request: nodejs-grunt-wrap - A Grunt plugin for wrapping project '
                      'text files')}],
        'builds': [{
            'epoch': 0,
            'nvr': 'nodejs-grunt-wrap-0.3.0-2.fc25',
            'signed': True}],
        'close_bugs': True,
        'comments': [
            {'bug_feedback': [],
             'id': 561418,
             'karma': 0,
             'karma_critpath': 0,
             'testcase_feedback': [],
             'text': 'This update has been submitted for testing by bowlofeggs. ',
             'timestamp': '2017-02-13 17:38:43',
             'update_id': 79733,
             'user': {'avatar': 'https://apps.fedoraproject.org/img/icons/bodhi-24.png',
                      'email': None,
                      'groups': [],
                      'id': 91,
                      'name': 'bodhi',
                      'openid': 'bodhi.id.fedoraproject.org'},
             'user_id': 91},
            {'bug_feedback': [],
             'id': 561619,
             'karma': 0,
             'karma_critpath': 0,
             'testcase_feedback': [],
             'text': 'This update has been pushed to testing.',
             'timestamp': '2017-02-14 00:55:18',
             'update_id': 79733,
             'user': {'avatar': 'https://apps.fedoraproject.org/img/icons/bodhi-24.png',
                      'email': None,
                      'groups': [],
                      'id': 91,
                      'name': 'bodhi',
                      'openid': 'bodhi.id.fedoraproject.org'},
             'user_id': 91},
            {'bug_feedback': [],
             'id': 562620,
             'karma': 0,
             'karma_critpath': 0,
             'testcase_feedback': [],
             'text': 'i found $100',
             'timestamp': '2017-02-28 14:47:43',
             'update_id': 79733,
             'user': {'avatar': 'AVATAR_URL',
                      'email': 'email@example.com',
                      'groups': [{'name': 'packager'}],
                      'id': 2897,
                      'name': 'bowlofeggs',
                      'openid': 'bowlofeggs.id.fedoraproject.org'},
             'user_id': 2897}],
        'critpath': False,
        'date_approved': None,
        'date_locked': None,
        'date_modified': None,
        'date_pushed': '2017-02-13 22:56:06',
        'date_stable': None,
        'date_submitted': '2017-02-13 17:38:43',
        'date_testing': '2017-02-13 22:56:06',
        'karma': 0,
        'locked': False,
        'meets_testing_requirements': True,
        'notes': 'New package.',
        'pushed': True,
        'release': {'branch': 'f25',
                    'candidate_tag': 'f25-updates-candidate',
                    'dist_tag': 'f25',
                    'id_prefix': 'FEDORA',
                    'long_name': 'Fedora 25',
                    'name': 'F25',
                    'override_tag': 'f25-override',
                    'pending_signing_tag': 'f25-signing-pending',
                    'pending_stable_tag': 'f25-updates-pending',
                    'pending_testing_tag': 'f25-updates-testing-pending',
                    'stable_tag': 'f25-updates',
                    'state': 'current',
                    'testing_tag': 'f25-updates-testing',
                    'version': '25',
                    'package_manager': 'unspecified',
                    'testing_repository': None,
                    'released_on': None,
                    'eol': None,
                    'setting_status': None},
        'request': None,
        'require_bugs': False,
        'require_testcases': False,
        'requirements': '',
        'severity': 'low',
        'stable_karma': 3,
        'status': 'testing',
        'submitter': 'bowlofeggs',
        'suggest': 'unspecified',
        'test_cases': [],
        'title': 'nodejs-grunt-wrap-0.3.0-2.fc25',
        'type': 'newpackage',
        'unstable_karma': -3,
        'updateid': 'FEDORA-2017-c95b33872d',
        'url': 'http://localhost:6543/updates/FEDORA-2017-c95b33872d',
        'user': {'avatar': 'AVATAR_URL',
                 'email': 'email@example.com',
                 'groups': [{'name': 'packager'}],
                 'id': 2897,
                 'name': 'bowlofeggs',
                 'openid': 'bowlofeggs.id.fedoraproject.org'}}]})

EXPECTED_QUERY_OUTPUT = """\
================================================================================
     nodejs-grunt-wrap-0.3.0-2.fc25
================================================================================
   Update ID: FEDORA-2017-c95b33872d
Content Type: rpm
     Release: Fedora 25
      Status: testing
        Type: newpackage
    Severity: low
       Karma: 0
   Autokarma: True  [-3, 3]
    Autotime: True
        Bugs: 1420605 - Review Request: nodejs-grunt-wrap - A Grunt plugin for
            : wrapping project text files
       Notes: New package.
   Submitter: bowlofeggs
   Submitted: 2017-02-13 17:38:43
    Comments: bodhi - 2017-02-13 17:38:43 (karma 0)
              This update has been submitted for testing by bowlofeggs.
              bodhi - 2017-02-14 00:55:18 (karma 0)
              This update has been pushed to testing.
              bowlofeggs - 2017-02-28 14:47:43 (karma 0)
              i found $100

  http://localhost:6543/updates/FEDORA-2017-c95b33872d

1 updates found (1 shown)"""

EXAMPLE_QUERY_MUNCH_MULTI = Munch({
    'chrome': True,
    'display_request': True,
    'display_user': True,
    'package': None,
    'page': 1,
    'pages': 1,
    'rows_per_page': 20,
    'total': 2,
    'updates': [{
        'alias': 'FEDORA-2017-c95b33872d',
        'autokarma': True,
        'autotime': True,
        'content_type': 'rpm',
        'bugs': [{
            'bug_id': 1420605,
            'feedback': [],
            'parent': False,
            'security': False,
            'title': ('Review Request: nodejs-grunt-wrap - A Grunt plugin for wrapping project '
                      'text files')}],
        'builds': [{
            'epoch': 0,
            'nvr': 'nodejs-grunt-wrap-0.3.0-2.fc25',
            'signed': True}],
        'close_bugs': True,
        'comments': [],
        'critpath': False,
        'date_approved': None,
        'date_locked': None,
        'date_modified': None,
        'date_pushed': '2017-02-13 22:56:06',
        'date_stable': None,
        'date_submitted': '2017-02-13 17:38:43',
        'date_testing': '2017-02-13 22:56:06',
        'karma': 0,
        'locked': False,
        'meets_testing_requirements': True,
        'notes': 'New package.',
        'pushed': True,
        'release': {'branch': 'f25',
                    'candidate_tag': 'f25-updates-candidate',
                    'dist_tag': 'f25',
                    'id_prefix': 'FEDORA',
                    'long_name': 'Fedora 25',
                    'name': 'F25',
                    'override_tag': 'f25-override',
                    'pending_signing_tag': 'f25-signing-pending',
                    'pending_stable_tag': 'f25-updates-pending',
                    'pending_testing_tag': 'f25-updates-testing-pending',
                    'stable_tag': 'f25-updates',
                    'state': 'current',
                    'testing_tag': 'f25-updates-testing',
                    'version': '25',
                    'package_manager': 'unspecified',
                    'testing_repository': None,
                    'released_on': None,
                    'eol': None,
                    'setting_status': None},
        'request': None,
        'require_bugs': False,
        'require_testcases': False,
        'requirements': '',
        'severity': 'low',
        'stable_karma': 3,
        'status': 'testing',
        'submitter': 'bowlofeggs',
        'suggest': 'unspecified',
        'test_cases': [],
        'title': 'nodejs-grunt-wrap-0.3.0-2.fc25',
        'type': 'newpackage',
        'unstable_karma': -3,
        'updateid': 'FEDORA-2017-c95b33872d',
        'url': 'http://localhost:6543/updates/FEDORA-2017-c95b33872d',
        'user': {'avatar': 'AVATAR_URL',
                 'email': 'email@example.com',
                 'groups': [{'name': 'packager'}],
                 'id': 2897,
                 'name': 'bowlofeggs',
                 'openid': 'bowlofeggs.id.fedoraproject.org'}},
        {
        'alias': 'FEDORA-2017-c95b33872d',
        'autokarma': True,
        'autotime': True,
        'content_type': 'rpm',
        'bugs': [{
            'bug_id': 1420605,
            'feedback': [],
            'parent': False,
            'security': False,
            'title': ('Review Request: nodejs-grunt-wrap - A Grunt plugin for wrapping project '
                      'text files')}],
        'builds': [{
            'epoch': 0,
            'nvr': 'nodejs-grunt-wrap-0.3.0-2.fc25',
            'signed': True}],
        'close_bugs': True,
        'comments': [],
        'critpath': False,
        'date_approved': None,
        'date_locked': None,
        'date_modified': None,
        'date_pushed': '2017-02-13 22:56:06',
        'date_stable': None,
        'date_submitted': '2017-02-13 17:38:43',
        'date_testing': '2017-02-13 22:56:06',
        'karma': 0,
        'locked': False,
        'meets_testing_requirements': True,
        'notes': 'New package.',
        'pushed': True,
        'release': {'branch': 'f25',
                    'candidate_tag': 'f25-updates-candidate',
                    'dist_tag': 'f25',
                    'id_prefix': 'FEDORA',
                    'long_name': 'Fedora 25',
                    'name': 'F25',
                    'override_tag': 'f25-override',
                    'pending_signing_tag': 'f25-signing-pending',
                    'pending_stable_tag': 'f25-updates-pending',
                    'pending_testing_tag': 'f25-updates-testing-pending',
                    'stable_tag': 'f25-updates',
                    'state': 'current',
                    'testing_tag': 'f25-updates-testing',
                    'version': '25',
                    'package_manager': 'unspecified',
                    'testing_repository': None,
                    'released_on': None,
                    'eol': None,
                    'setting_status': None},
        'request': None,
        'require_bugs': False,
        'require_testcases': False,
        'requirements': '',
        'severity': 'low',
        'stable_karma': 3,
        'status': 'testing',
        'submitter': 'bowlofeggs',
        'suggest': 'unspecified',
        'test_cases': [],
        'title': 'nodejs-grunt-wrap-0.3.0-2.fc25',
        'type': 'newpackage',
        'unstable_karma': -3,
        'updateid': 'FEDORA-2017-c95b33872d',
        'url': 'http://localhost:6543/updates/FEDORA-2017-c95b33872d',
        'user': {'avatar': 'AVATAR_URL',
                  'email': 'email@example.com',
                  'groups': [{'name': 'packager'}],
                  'id': 2897,
                  'name': 'bowlofeggs',
                  'openid': 'bowlofeggs.id.fedoraproject.org'}}]})

EXAMPLE_QUERY_OUTPUT_MULTI = """\
 nodejs-grunt-wrap-0.3.0-2.fc25           rpm        testing   2017-02-13 (17)
 nodejs-grunt-wrap-0.3.0-2.fc25           rpm        testing   2017-02-13 (17)
2 updates found (2 shown)
"""

EXAMPLE_QUERY_OVERRIDES_MUNCH = Munch({
    'chrome': True,
    'display_user': True,
    'overrides': [
        {'build': {
            'epoch': 0,
            'nvr': 'nodejs-grunt-wrap-0.3.0-2.fc25',
            'signed': True},
         'build_id': 108565,
         'expiration_date': '2017-03-07 14:30:36',
         'expired_date': None,
         'notes': 'No explanation given...',
         'nvr': 'nodejs-grunt-wrap-0.3.0-2.fc25',
         'submission_date': '2017-02-28 14:30:37',
         'submitter': {'avatar': 'AVATAR_URL',
                       'email': 'email@example.com',
                       'groups': [{'name': 'packager'}],
                       'id': 2897,
                       'name': 'bowlofeggs',
                       'openid': 'bowlofeggs.id.fedoraproject.org'},
         'submitter_id': 2897},
        {'build': {'epoch': 0,
                   'nvr': 'python-pyramid-1.5.6-3.el7',
                   'signed': True},
         'build_id': 107673,
         'expiration_date': '2017-02-17 00:00:00',
         'expired_date': None,
         'notes': 'This is needed to build bodhi-2.4.0.',
         'nvr': 'python-pyramid-1.5.6-3.el7',
         'submission_date': '2017-02-03 20:08:46',
         'submitter': {'avatar': 'AVATAR_URL',
                       'email': 'email@example.com',
                       'groups': [{'name': 'packager'}],
                       'id': 2897,
                       'name': 'bowlofeggs',
                       'openid': 'bowlofeggs.id.fedoraproject.org'},
         'submitter_id': 2897},
        {'build': {'epoch': 0,
                   'nvr': 'erlang-esip-1.0.8-1.fc25',
                   'signed': True},
         'build_id': 98946,
         'expiration_date': '2016-11-12 16:59:29',
         'expired_date': '2016-11-12 17:00:04',
         'notes': 'needed for ejabberd',
         'nvr': 'erlang-esip-1.0.8-1.fc25',
         'submission_date': '2016-11-10 16:59:35',
         'submitter': {'avatar': 'AVATAR_URL',
                       'email': 'email@example.com',
                       'groups': [{'name': 'packager'}],
                       'id': 2897,
                       'name': 'bowlofeggs',
                       'openid': 'bowlofeggs.id.fedoraproject.org'},
         'submitter_id': 2897},
        {'build': {'epoch': 0,
                   'nvr': 'erlang-stun-1.0.7-1.fc25',
                   'signed': True},
         'build_id': 98945,
         'expiration_date': '2016-11-12 00:00:00',
         'expired_date': '2016-11-12 00:00:22',
         'notes': 'This is needed for ejabberd.',
         'nvr': 'erlang-stun-1.0.7-1.fc25',
         'submission_date': '2016-11-10 16:21:53',
         'submitter': {'avatar': 'AVATAR_URL',
                       'email': 'email@example.com',
                       'groups': [{'name': 'packager'}],
                       'id': 2897,
                       'name': 'bowlofeggs',
                       'openid': 'bowlofeggs.id.fedoraproject.org'},
         'submitter_id': 2897},
        {'build': {'epoch': 0,
                   'nvr': 'erlang-iconv-1.0.2-1.fc25',
                   'signed': True},
         'build_id': 98942,
         'expiration_date': '2016-11-12 00:00:00',
         'expired_date': '2016-11-12 00:00:21',
         'notes': 'This is needed for ejabberd.',
         'nvr': 'erlang-iconv-1.0.2-1.fc25',
         'submission_date': '2016-11-10 15:45:17',
         'submitter': {'avatar': 'AVATAR_URL',
                       'email': 'email@example.com',
                       'groups': [{'name': 'packager'}],
                       'id': 2897,
                       'name': 'bowlofeggs',
                       'openid': 'bowlofeggs.id.fedoraproject.org'},
         'submitter_id': 2897},
        {'build': {'epoch': 0,
                   'nvr': 'erlang-stringprep-1.0.6-1.fc25',
                   'signed': True},
         'build_id': 98941,
         'expiration_date': '2016-11-12 00:00:00',
         'expired_date': '2016-11-12 00:00:24',
         'notes': 'This is needed for ejabberd.',
         'nvr': 'erlang-stringprep-1.0.6-1.fc25',
         'submission_date': '2016-11-10 15:43:52',
         'submitter': {'avatar': 'AVATAR_URL',
                       'email': 'email@example.com',
                       'groups': [{'name': 'packager'}],
                       'id': 2897,
                       'name': 'bowlofeggs',
                       'openid': 'bowlofeggs.id.fedoraproject.org'},
         'submitter_id': 2897},
        {'build': {'epoch': 0,
                   'nvr': 'erlang-fast_tls-1.0.7-1.fc25',
                   'signed': True},
         'build_id': 98940,
         'expiration_date': '2016-11-12 00:00:00',
         'expired_date': '2016-11-12 00:00:26',
         'notes': 'This is needed for ejabberd.',
         'nvr': 'erlang-fast_tls-1.0.7-1.fc25',
         'submission_date': '2016-11-10 15:41:09',
         'submitter': {'avatar': 'AVATAR_URL',
                       'email': 'email@example.com',
                       'groups': [{'name': 'packager'}],
                       'id': 2897,
                       'name': 'bowlofeggs',
                       'openid': 'bowlofeggs.id.fedoraproject.org'},
         'submitter_id': 2897},
        {'build': {'epoch': 0,
                   'nvr': 'erlang-fast_yaml-1.0.6-1.fc25',
                   'signed': True},
         'build_id': 98939,
         'expiration_date': '2016-11-12 00:00:00',
         'expired_date': '2016-11-12 00:00:27',
         'notes': 'This is needed for ejabberd.',
         'nvr': 'erlang-fast_yaml-1.0.6-1.fc25',
         'submission_date': '2016-11-10 15:39:25',
         'submitter': {'avatar': 'AVATAR_URL',
                       'email': 'email@example.com',
                       'groups': [{'name': 'packager'}],
                       'id': 2897,
                       'name': 'bowlofeggs',
                       'openid': 'bowlofeggs.id.fedoraproject.org'},
         'submitter_id': 2897},
        {'build': {'epoch': 0,
                   'nvr': 'erlang-fast_xml-1.1.15-1.fc25',
                   'signed': True},
         'build_id': 98938,
         'expiration_date': '2016-11-12 15:30:10',
         'expired_date': '2016-11-12 16:00:04',
         'notes': 'needed for ejabberd',
         'nvr': 'erlang-fast_xml-1.1.15-1.fc25',
         'submission_date': '2016-11-10 15:30:16',
         'submitter': {'avatar': 'AVATAR_URL',
                       'email': 'email@example.com',
                       'groups': [{'name': 'packager'}],
                       'id': 2897,
                       'name': 'bowlofeggs',
                       'openid': 'bowlofeggs.id.fedoraproject.org'},
         'submitter_id': 2897},
        {'build': {'epoch': 0,
                   'nvr': 'python-fedmsg-atomic-composer-2016.3-1.el7',
                   'signed': True},
         'build_id': 97312,
         'expiration_date': '2017-02-17 00:00:00',
         'expired_date': None,
         'notes': 'This is needed to build bodhi-2.4.0.',
         'nvr': 'python-fedmsg-atomic-composer-2016.3-1.el7',
         'submission_date': '2016-10-27 15:55:34',
         'submitter': {'avatar': 'AVATAR_URL',
                       'email': 'email@example.com',
                       'groups': [{'name': 'packager'}],
                       'id': 2897,
                       'name': 'bowlofeggs',
                       'openid': 'bowlofeggs.id.fedoraproject.org'},
         'submitter_id': 2897},
        {'build': {'epoch': 0,
                   'nvr': 'python-fedmsg-atomic-composer-2016.3-1.fc24',
                   'signed': True},
         'build_id': 97311,
         'expiration_date': '2016-10-29 00:00:00',
         'expired_date': '2016-10-29 00:00:23',
         'notes': 'This is needed to build bodhi-2.3.0.',
         'nvr': 'python-fedmsg-atomic-composer-2016.3-1.fc24',
         'submission_date': '2016-10-27 15:50:43',
         'submitter': {'avatar': 'AVATAR_URL',
                       'email': 'email@example.com',
                       'groups': [{'name': 'packager'}],
                       'id': 2897,
                       'name': 'bowlofeggs',
                       'openid': 'bowlofeggs.id.fedoraproject.org'},
         'submitter_id': 2897}],
    'page': 1,
    'pages': 1,
    'rows_per_page': 20,
    'total': 11})


EXAMPLE_QUERY_SINGLE_OVERRIDE_MUNCH = Munch({
    'chrome': True,
    'display_user': True,
    'overrides': [Munch(
        {'build': Munch({
            'epoch': 0,
            'nvr': 'js-tag-it-2.0-1.fc25',
            'release_id': 15,
            'signed': True}),
         'build_id': 108565,
         'expiration_date': '2017-03-07 23:05:31',
         'expired_date': None,
         'notes': 'No explanation given...',
         'nvr': 'nodejs-grunt-wrap-0.3.0-2.fc25',
         'submission_date': '2017-02-28 14:30:37',
         'submitter': {'avatar': 'AVATAR_URL',
                       'email': 'email@example.com',
                       'groups': [{'name': 'packager'}],
                       'id': 2897,
                       'name': 'bowlofeggs',
                       'openid': 'bowlofeggs.id.fedoraproject.org'},
         'submitter_id': 2897})],
    'page': 1,
    'pages': 1,
    'rows_per_page': 20,
    'total': 1})

# Expected output when print_resp renders EXAMPLE_QUERY_OVERRIDES_MUNCH
EXPECTED_QUERY_OVERRIDES_OUTPUT = (
    """bowlofeggs's nodejs-grunt-wrap-0.3.0-2.fc25 override (expires 2017-03-07 14:30:36)
bowlofeggs's python-pyramid-1.5.6-3.el7 override (expires 2017-02-17 00:00:00)
bowlofeggs's erlang-esip-1.0.8-1.fc25 override (expires 2016-11-12 16:59:29)
bowlofeggs's erlang-stun-1.0.7-1.fc25 override (expires 2016-11-12 00:00:00)
bowlofeggs's erlang-iconv-1.0.2-1.fc25 override (expires 2016-11-12 00:00:00)
bowlofeggs's erlang-stringprep-1.0.6-1.fc25 override (expires 2016-11-12 00:00:00)
bowlofeggs's erlang-fast_tls-1.0.7-1.fc25 override (expires 2016-11-12 00:00:00)
bowlofeggs's erlang-fast_yaml-1.0.6-1.fc25 override (expires 2016-11-12 00:00:00)
bowlofeggs's erlang-fast_xml-1.1.15-1.fc25 override (expires 2016-11-12 15:30:10)
bowlofeggs's python-fedmsg-atomic-composer-2016.3-1.el7 override (expires 2017-02-17 00:00:00)
bowlofeggs's python-fedmsg-atomic-composer-2016.3-1.fc24 override (expires 2016-10-29 00:00:00)
11 overrides found (11 shown)
""")


EXAMPLE_UPDATE_MUNCH = Munch({
    'date_testing': '2016-10-06 00:55:15', 'pushed': True,
    'require_testcases': True, 'date_locked': None, 'critpath': False, 'date_approved': None,
    'stable_karma': 3, 'date_pushed': '2016-10-21 13:23:01', 'requirements': '',
    'severity': 'unspecified', 'autokarma': True, 'autotime': True, 'title': 'bodhi-2.2.4-1.el7',
    'suggest': 'unspecified', 'require_bugs': True,
    'comments': [
        Munch({
            'bug_feedback': [], 'user_id': 91, 'timestamp': '2016-10-05 18:10:22',
            'testcase_feedback': [], 'karma_critpath': 0, 'update_id': 69704, 'karma': 0,
            'text': 'This update has been submitted for testing by bowlofeggs. ', 'id': 501425,
            'user': Munch({
                'openid': 'bodhi.id.fedoraproject.org', 'name': 'bodhi',
                'id': 91, 'avatar': 'https://apps.fedoraproject.org/img/icons/bodhi-24.png',
                'groups': [], 'email': None})}),
        Munch({
            'bug_feedback': [], 'user_id': 91, 'timestamp': '2016-10-05 18:10:27',
            'testcase_feedback': [], 'karma_critpath': 0, 'update_id': 69704, 'karma': 0,
            'text': ('This update has obsoleted [bodhi-2.2.3-1.el7]'
                     '(https://bodhi.fedoraproject.org/updates/FEDORA-EPEL-2016-a0eb4cc41f), and '
                     'has inherited its bugs and notes.'),
            'id': 501427,
            'user': Munch({
                'openid': 'bodhi.id.fedoraproject.org', 'name': 'bodhi',
                'id': 91, 'avatar': 'https://apps.fedoraproject.org/img/icons/bodhi-24.png',
                'groups': [], 'email': None})})],
    'updateid': 'FEDORA-EPEL-2016-3081a94111', 'type': 'bugfix', 'close_bugs': True,
    'meets_testing_requirements': True, 'date_submitted': '2016-10-05 18:10:22',
    'unstable_karma': -3, 'submitter': 'bowlofeggs',
    'user': Munch({
        'openid': 'bowlofeggs.id.fedoraproject.org', 'name': 'bowlofeggs',
        'id': 2897,
        'avatar': 'https://seccdn.libravatar.org/avatar/some_hash',
        'groups': [Munch({'name': 'packager'})], 'email': 'bowlofeggs@electronsweatshop.com'}),
    'locked': False,
    'builds': [Munch({'epoch': 0, 'nvr': 'bodhi-2.2.4-1.el7', 'signed': True})],
    'date_modified': None,
    'url': 'https://bodhi.fedoraproject.org/updates/FEDORA-EPEL-2016-3081a94111',
    'test_cases': [],
    'notes': ('Update to 2.2.4. Release notes available at '
              'https://github.com/fedora-infra/bodhi/releases/tag/2.2.4\n'),
    'request': None, 'bugs': [], 'alias': 'FEDORA-EPEL-2016-3081a94111', 'status': 'stable',
    'karma': 0,
    'release': Munch({
        'dist_tag': 'epel7', 'name': 'EPEL-7', 'testing_tag': 'epel7-testing',
        'pending_stable_tag': 'epel7-pending', 'pending_signing_tag': 'epel7-signing-pending',
        'long_name': 'Fedora EPEL 7', 'state': 'current', 'version': '7',
        'override_tag': 'epel7-override', 'branch': 'epel7', 'id_prefix': 'FEDORA-EPEL',
        'pending_testing_tag': 'epel7-testing-pending', 'stable_tag': 'epel7',
        'candidate_tag': 'epel7-testing-candidate', 'package_manager': 'unspecified',
        'testing_repository': None, 'released_on': None, 'eol': None,
        'setting_status': None}), 'date_stable': '2016-10-21 13:23:01',
    'content_type': 'rpm'})

SINGLE_UPDATE_MUNCH = Munch({
    'update': Munch({
        'date_testing': '2016-10-06 00:55:15', 'pushed': True,
        'require_testcases': True, 'date_locked': None,
        'critpath': False, 'date_approved': None,
        'stable_karma': 3, 'date_pushed': '2016-10-21 13:23:01', 'requirements': '',
        'severity': 'unspecified', 'autokarma': True, 'autotime': True,
        'title': 'bodhi-2.2.4-1.el7', 'suggest': 'unspecified', 'require_bugs': True,
        'comments': [
            Munch({
                'bug_feedback': [], 'user_id': 91, 'timestamp': '2016-10-05 18:10:22',
                'testcase_feedback': [], 'karma_critpath': 0, 'update_id': 69704, 'karma': 0,
                'text': 'This update has been submitted for testing by bowlofeggs. ',
                'id': 501425,
                'user': Munch({
                    'openid': 'bodhi.id.fedoraproject.org', 'name': 'bodhi',
                    'id': 91, 'avatar': 'https://apps.fedoraproject.org/img/icons/bodhi-24.png',
                    'groups': [], 'email': None})}),
            Munch({
                'bug_feedback': [], 'user_id': 91, 'timestamp': '2016-10-05 18:10:27',
                'testcase_feedback': [], 'karma_critpath': 0, 'update_id': 69704, 'karma': 0,
                'text': ('This update has obsoleted [bodhi-2.2.3-1.el7]'
                         '(https://bodhi.fedoraproject.org/updates/FEDORA-EPEL-2016-a0eb4cc41f), '
                         'and has inherited its bugs and notes.'),
                'id': 501427,
                'user': Munch({
                    'openid': 'bodhi.id.fedoraproject.org', 'name': 'bodhi',
                    'id': 91, 'avatar': 'https://apps.fedoraproject.org/img/icons/bodhi-24.png',
                    'groups': [], 'email': None})})],
        'updateid': 'FEDORA-EPEL-2016-3081a94111', 'type': 'bugfix', 'close_bugs': True,
        'meets_testing_requirements': True, 'date_submitted': '2016-10-05 18:10:22',
        'unstable_karma': -3, 'submitter': 'bowlofeggs',
        'user': Munch({
            'openid': 'bowlofeggs.id.fedoraproject.org', 'name': 'bowlofeggs',
            'id': 2897,
            'avatar': 'https://seccdn.libravatar.org/avatar/some_hash',
            'groups': [Munch({'name': 'packager'})],
            'email': 'bowlofeggs@electronsweatshop.com'}),
        'locked': False,
        'builds': [Munch({'epoch': 0, 'nvr': 'bodhi-2.2.4-1.el7', 'signed': True})],
        'date_modified': None,
        'url': 'https://bodhi.fedoraproject.org/updates/FEDORA-EPEL-2016-3081a94111',
        'test_cases': [],
        'notes': ('Update to 2.2.4. Release notes available at '
                  'https://github.com/fedora-infra/bodhi/releases/tag/2.2.4\n'),
        'request': None, 'bugs': [], 'alias': 'FEDORA-EPEL-2016-3081a94111',
        'status': 'stable', 'karma': 0,
        'release': Munch({
            'dist_tag': 'epel7', 'name': 'EPEL-7', 'testing_tag': 'epel7-testing',
            'pending_stable_tag': 'epel7-pending',
            'pending_signing_tag': 'epel7-signing-pending',
            'long_name': 'Fedora EPEL 7', 'state': 'current', 'version': '7',
            'override_tag': 'epel7-override', 'branch': 'epel7', 'id_prefix': 'FEDORA-EPEL',
            'pending_testing_tag': 'epel7-testing-pending', 'stable_tag': 'epel7',
            'candidate_tag': 'epel7-testing-candidate', 'package_manager': 'unspecified',
            'testing_repository': None, 'released_on': None, 'eol': None,
            'setting_status': None}), 'date_stable': '2016-10-21 13:23:01',
        'content_type': 'rpm'})})


EXAMPLE_GET_RELEASE_15 = Munch(
    {'rows_per_page': 20, 'total': 1, 'pages': 1,
     'releases': [
         Munch({'dist_tag': 'f25', 'name': 'F25', 'testing_tag': 'f25-updates-testing',
                'pending_stable_tag': 'f25-updates-pending',
                'pending_signing_tag': 'f25-signing-pending', 'long_name': 'Fedora 25',
                'state': 'current', 'version': '25', 'override_tag': 'f25-override',
                'branch': 'f25', 'id_prefix': 'FEDORA',
                'pending_testing_tag': 'f25-updates-testing-pending',
                'stable_tag': 'f25-updates', 'candidate_tag': 'f25-updates-candidate',
                'package_manager': 'unspecified', 'testing_repository': None, 'released_on': None,
                'eol': None, 'setting_status': None})],
     'page': 1})


# EXAMPLE_UPDATE_MUNCH is expected to generate this output in update_str
EXPECTED_UPDATE_OUTPUT = (
    """================================================================================
     bodhi-2.2.4-1.el7
================================================================================
   Update ID: FEDORA-EPEL-2016-3081a94111
Content Type: rpm
     Release: Fedora EPEL 7
      Status: stable
        Type: bugfix
    Severity: unspecified
       Karma: 0
   Autokarma: True  [-3, 3]
    Autotime: True
       Notes: Update to 2.2.4. Release notes available at https://github.com
            : /fedora-infra/bodhi/releases/tag/2.2.4
   Submitter: bowlofeggs
   Submitted: 2016-10-05 18:10:22
    Comments: bodhi - 2016-10-05 18:10:22 (karma 0)
              This update has been submitted for testing by bowlofeggs.
              bodhi - 2016-10-05 18:10:27 (karma 0)
              This update has obsoleted
              [bodhi-2.2.3-1.el7](https://bodhi.fedoraproject.org/updates
              /FEDORA-EPEL-2016-a0eb4cc41f), and has inherited its bugs and
              notes.

  http://example.com/tests/updates/FEDORA-EPEL-2016-3081a94111
""")

EXPECTED_OVERRIDE_STR_OUTPUT = """============================================================
     js-tag-it-2.0-1.fc25
============================================================
  Submitter: bowlofeggs
  Expiration Date: 2017-03-07 23:05:31
  Notes: No explanation given...
  Expired: False
"""


EXPECTED_OVERRIDES_OUTPUT = EXPECTED_OVERRIDE_STR_OUTPUT + """

Use the following to ensure the override is active:

\t$ koji wait-repo f25-build --build=js-tag-it-2.0-1.fc25

"""

EXPECTED_EXPIRED_OVERRIDES_OUTPUT = """============================================================
     js-tag-it-2.0-1.fc25
============================================================
  Submitter: bowlofeggs
  Expiration Date: 2017-03-07 23:05:31
  Notes: This is an expired override
  Expired: True
"""

EXAMPLE_RELEASE_MUNCH = Munch({
    'dist_tag': 'f27', 'testing_tag': 'f27-updates-testing', 'branch': 'f27',
    'pending_stable_tag': 'f27-updates-pending', 'pending_signing_tag': 'f27-signing-pending',
    'long_name': 'Fedora 27', 'state': 'pending', 'version': '27', 'name': 'F27',
    'override_tag': 'f27-override', 'id_prefix': 'FEDORA', 'composed_by_bodhi': True,
    'pending_testing_tag': 'f27-updates-testing-pending', 'stable_tag': 'f27-updates',
    'candidate_tag': 'f27-updates-candidate', 'mail_template': 'fedora_errata_template',
    'create_automatic_updates': False, 'package_manager': 'unspecified',
    'testing_repository': None, 'released_on': None, 'eol': None, 'setting_status': None})


EXPECTED_RELEASE_OUTPUT = """Saved release:
  Name:                     F27
  Long Name:                Fedora 27
  Version:                  27
  Branch:                   f27
  ID Prefix:                FEDORA
  Dist Tag:                 f27
  Stable Tag:               f27-updates
  Testing Tag:              f27-updates-testing
  Candidate Tag:            f27-updates-candidate
  Pending Signing Tag:      f27-signing-pending
  Pending Testing Tag:      f27-updates-testing-pending
  Pending Stable Tag:       f27-updates-pending
  Override Tag:             f27-override
  State:                    pending
  Beta Status:              None
  Email Template:           fedora_errata_template
  Composed by Bodhi:        True
  Create Automatic Updates: False
  Package Manager:          unspecified
  Testing Repository:       None
  Released On:              None
  End of Life:              None
"""

EXAMPLE_ARCHIVED_RELEASE_MUNCH = Munch({
    'composes': [], 'dist_tag': 'f26', 'name': 'F26',
    'testing_tag': 'f26-updates-testing', 'pending_stable_tag': 'f26-updates-pending',
    'mail_template': 'fedora_errata_template', 'long_name': 'Fedora 26',
    'state': 'archived', 'version': '26', 'id_prefix': 'FEDORA', 'branch': 'f26',
    'pending_signing_tag': 'f26-signing-pending',
    'pending_testing_tag': 'f26-updates-testing-pending',
    'candidate_tag': 'f26-updates-candidate', 'stable_tag': 'f26-updates',
    'override_tag': 'f26-override', 'composed_by_bodhi': True,
    'package_manager': 'unspecified', 'testing_repository': None, 'released_on': None,
    'eol': None, 'setting_status': None,
})

EXAMPLE_CURRENT_RELEASE_MUNCH = Munch({
    'composes': [], 'dist_tag': 'f28', 'name': 'F28',
    'testing_tag': 'f28-updates-testing', 'pending_stable_tag': 'f28-updates-pending',
    'mail_template': 'fedora_errata_template', 'long_name': 'Fedora 28',
    'state': 'current', 'version': '28', 'id_prefix': 'FEDORA', 'branch': 'f28',
    'pending_signing_tag': 'f28-signing-pending',
    'pending_testing_tag': 'f28-updates-testing-pending',
    'candidate_tag': 'f28-updates-candidate', 'stable_tag': 'f28-updates',
    'override_tag': 'f28-override', 'composed_by_bodhi': True,
    'package_manager': 'unspecified', 'testing_repository': None, 'released_on': None,
    'eol': None, 'setting_status': None,
})

EXAMPLE_PENDING_RELEASE_MUNCH = Munch({
    'composes': [], 'dist_tag': 'f29', 'name': 'F29',
    'testing_tag': 'f29-updates-testing', 'pending_stable_tag': 'f29-updates-pending',
    'mail_template': 'fedora_errata_template', 'long_name': 'Fedora 29',
    'state': 'pending', 'version': '29', 'id_prefix': 'FEDORA', 'branch': 'f29',
    'pending_signing_tag': 'f29-signing-pending',
    'pending_testing_tag': 'f29-updates-testing-pending',
    'candidate_tag': 'f29-updates-candidate', 'stable_tag': 'f29-updates',
    'override_tag': 'f29-override', 'composed_by_bodhi': True,
    'package_manager': 'unspecified', 'testing_repository': None, 'released_on': None,
    'eol': None, 'setting_status': None,
})

EXAMPLE_FROZEN_RELEASE_MUNCH = Munch({
    'composes': [], 'dist_tag': 'f30', 'name': 'F30',
    'testing_tag': 'f30-updates-testing', 'pending_stable_tag': 'f30-updates-pending',
    'mail_template': 'fedora_errata_template', 'long_name': 'Fedora 30',
    'state': 'frozen', 'version': '30', 'id_prefix': 'FEDORA', 'branch': 'f30',
    'pending_signing_tag': 'f30-signing-pending',
    'pending_testing_tag': 'f30-updates-testing-pending',
    'candidate_tag': 'f30-updates-candidate', 'stable_tag': 'f30-updates',
    'override_tag': 'f30-override', 'composed_by_bodhi': True,
    'package_manager': 'unspecified', 'testing_repository': None, 'released_on': None,
    'eol': None, 'setting_status': None,
})

EXAMPLE_RELEASE_MUNCH_NO_ARCHIVED = Munch({
    'rows_per_page': 20, 'total': 3, 'pages': 1,
    'releases': [
        EXAMPLE_CURRENT_RELEASE_MUNCH,
        EXAMPLE_PENDING_RELEASE_MUNCH,
        EXAMPLE_FROZEN_RELEASE_MUNCH,
    ]
})

EXAMPLE_RELEASE_MUNCH_WITH_ARCHIVED = Munch({
    'rows_per_page': 20, 'total': 3, 'pages': 1,
    'releases': [
        EXAMPLE_ARCHIVED_RELEASE_MUNCH,
        EXAMPLE_CURRENT_RELEASE_MUNCH,
        EXAMPLE_PENDING_RELEASE_MUNCH,
    ]
})

EXPECTED_PENDING_RELEASES_LIST_OUTPUT = """pending:
  Name:                F29
"""

EXPECTED_ARCHIVED_RELEASES_LIST_OUTPUT = """archived:
  Name:                F26
"""

EXPECTED_CURRENT_RELEASES_LIST_OUTPUT = """current:
  Name:                F28
"""

EXPECTED_FROZEN_RELEASES_LIST_OUTPUT = """frozen:
  Name:                F30
"""


UNMATCHED_RESP = {"pants": "pants"}



# Copyright © 2018 Red Hat, Inc.
#
# This file is part of Bodhi.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""This module contains useful functions that helps with testing."""

from io import BytesIO
import difflib
import re

from requests import Response


def prepare_text_for_comparison(text):
    """Transfer wrapped text output to one line."""
    text = text.replace('\n', '')  # Remove newlines
    text = re.sub(r' +:', '', text)  # Remove start of wrapped lines
    text = "".join(text.split())  # Remove whitespaces
    return text


def compare_output(output, expected, debug_output=False):
    """Compare content of wrapped text outputs and show diff if enabled."""
    prepared_output = prepare_text_for_comparison(output)
    prepared_expected = prepare_text_for_comparison(expected)
    if prepared_output == prepared_expected:
        return True
    else:
        if debug_output:
            differ = difflib.Differ()
            diff = differ.compare(output.splitlines(True),
                                  expected.splitlines(True))
            print(''.join(list(diff)))
        return False


def build_response(status_code, url, content):
    """Build a response with the provided parameters.

    Args:
        status_code (int): The HTTP status code.
        url (str): The URL that was requested.
        content (str): The response content.

    Returns:
        requests.Response: A Response object compatible with ``requests``.
    """
    response = Response()
    response.status_code = status_code
    response.url = url
    response.encoding = "utf-8"
    response.raw = BytesIO(content.encode("utf-8"))
    return response



# bodhi documentation build configuration file, created by
# sphinx-quickstart on Sat Aug 10 09:29:50 2013.
#
# This file is execfile()d with the current directory set to its containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.
import datetime

# The suffix of source filenames.
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'bodhi'
copyright = '2007-{}, Red Hat, Inc.'.format(datetime.datetime.utcnow().year)

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build']

# -- Options for manual page output --------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ('man_pages/bodhi', 'bodhi', 'manage Fedora updates', ['Randy Barlow', 'Luke Macken'], 1),
]



"""Constants for the Bodhi Client."""

BASE_URL = "https://bodhi.fedoraproject.org/"
IDP = "https://id.fedoraproject.org/openidc"
CLIENT_ID = "bodhi-client"
STG_BASE_URL = "https://bodhi.stg.fedoraproject.org/"
STG_IDP = "https://id.stg.fedoraproject.org/openidc"
STG_CLIENT_ID = "bodhi-client"
SCOPE = " ".join([
    "openid",
    "email",
    "profile",
    "https://id.fedoraproject.org/scope/groups",
    "https://id.fedoraproject.org/scope/agreements",
])
UPDATE_TYPES = ['security', 'bugfix', 'enhancement', 'newpackage']
REQUEST_TYPES = ['testing', 'stable', 'unpush']
SUGGEST_TYPES = ['unspecified', 'reboot', 'logout']



# Copyright © 2007-2019 Red Hat, Inc. and others.
#
# This file is part of bodhi.
#
# This software is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this software; if not, see <http://www.gnu.org/licenses/>
"""
This module provides Python bindings to the Bodhi REST API.

.. moduleauthor:: Luke Macken <lmacken@redhat.com>
.. moduleauthor:: Toshio Kuratomi <tkuratom@redhat.com>
.. moduleauthor:: Ralph Bean <rbean@redhat.com>
.. moduleauthor:: Randy Barlow <bowlofeggs@fedoraproject.org>
"""

from urllib.parse import urlparse
import configparser
import datetime
import functools
import itertools
import logging
import os
import re
import textwrap
import typing


try:
    import dnf
except ImportError:  # pragma: no cover
    # dnf is not available on EL 7.
    dnf = None  # pragma: no cover
from munch import munchify
from requests.exceptions import ConnectionError, RequestException
import koji
import requests

from .constants import (
    BASE_URL,
    CLIENT_ID,
    IDP,
    SCOPE,
    STG_BASE_URL,
    STG_CLIENT_ID,
    STG_IDP,
)
from .oidcclient import JSONStorage, OIDCClient, OIDCClientError


if typing.TYPE_CHECKING:  # pragma: no cover
    import munch  # noqa: 401


log = logging.getLogger(__name__)


UPDATE_ID_RE = r'FEDORA-(EPEL-)?\d{4,4}'
UPDATE_TITLE_RE = r'(\.el|\.fc)\d\d?'


class BodhiClientException(RequestException):
    """Used to indicate there was an error in the client."""


class UpdateNotFound(BodhiClientException):
    """Used to indicate that a referenced Update is not found on the server."""

    def __init__(self, update: str, **kwargs):
        """
        Initialize the Exception.

        Args:
            update: The alias of the update that was not found.
        """
        super().__init__(**kwargs)
        self.update = update

    def __str__(self) -> str:
        """
        Return a human readable error message.

        Returns:
            An error message.
        """
        return f'Update not found: {self.update}'


class ComposeNotFound(BodhiClientException):
    """Used to indicate that a referenced Compose is not found on the server."""

    def __init__(self, release: str, request: str, **kwargs):
        """
        Initialize the Exception.

        Args:
            release: The release component of the compose that was not found.
            request: The request component of the compose that was not found.
        """
        super().__init__(**kwargs)
        self.release = release
        # self.request is already defined in parent class as type Request
        self.release_request = request

    def __str__(self) -> str:
        """
        Return a human readable error message.

        Returns:
            An error message.
        """
        return (f'Compose with request "{self.release_request}" '
                f'not found for release "{self.release}"')


def errorhandled(method: typing.Callable) -> typing.Callable:
    """Raise exceptions on failure. Used as a decorator for BodhiClient methods."""

    @functools.wraps(method)
    def wrapper(*args, **kwargs) -> typing.Any:
        result = method(*args, **kwargs)
        # Bodhi allows comments to be written by unauthenticated users if they solve a Captcha.
        # Due to this, an authentication error is not raised by the server if the client fails
        # to authenticate for any reason, and instead an error about needing a captcha key is
        # presented instead. If we see that error, we can just clear authentication and retry.
        if 'errors' in result:
            for error in result['errors']:
                if 'name' in error and error['name'] == 'captcha_key':
                    args[0].clear_auth()
                    result = method(*args, **kwargs)

        if 'errors' not in result:
            return result

        # Otherwise, there was a problem...
        problems = 'An unhandled error occurred in the BodhiClient'
        try:
            problems = "\n".join([e['description'] for e in result['errors']])
        except Exception:
            pass
        raise BodhiClientException(problems)
    return wrapper


def _days_since(data_str: str) -> int:
    """
    Return number of days since the datetime passed as input in the form '%Y-%m-%d %H:%M:%S'.

    This can be used to calculate how many days an update is in current state by passing
    directly the 'date_pushed' or 'date_submitted' from the Update object.
    This is also useful to easily mock the output, since datetime.datetime.utcnow()
    cannot be mocked.

    Args:
        data_str: The 'date_pushed' or 'date_submitted' from the Update object.

    Returns:
        Number of days since the date in input.
    """
    update_time = datetime.datetime.strptime(data_str, '%Y-%m-%d %H:%M:%S')
    return (datetime.datetime.utcnow() - update_time).days


class BodhiClient:
    """Python bindings to the Bodhi server REST API."""

    def __init__(
        self,
        base_url: str = BASE_URL,
        client_id: str = CLIENT_ID,
        id_provider: str = IDP,
        staging: bool = False,
        oidc_storage_path: typing.Optional[str] = None,
    ):
        """
        Initialize the Bodhi client.

        Args:
            base_url: The URL of the Bodhi server to connect to. Ignored if
                      ```staging``` is True.
            client_id: The OpenID Connect Client ID.
            staging: If True, use the staging server. If False, use base_url.
            oidc_storage_path: Path to a file were OIDC credentials are stored
        """
        if staging:
            base_url = STG_BASE_URL
            id_provider = STG_IDP
            client_id = STG_CLIENT_ID

        if base_url[-1] != '/':
            base_url = base_url + '/'
        self.base_url = base_url
        self.csrf_token = ''
        self.oidc_storage_path = (
            oidc_storage_path or os.path.join(
                os.path.expanduser("~"), ".config", "bodhi", "client.json"
            )
        )
        self._build_oidc_client(client_id, id_provider)

    def _build_oidc_client(self, client_id, id_provider):
        self.oidc = OIDCClient(
            client_id,
            SCOPE,
            id_provider.rstrip("/"),
            storage=JSONStorage(self.oidc_storage_path),
        )

    def send_request(self, url, verb="GET", **kwargs):
        """Send the request to the Bodhi server.

        Args:
            url (str): The relative URL on the Bodhi server
            verb (str, optional): The HTTP method. Defaults to "GET".

        Returns:
            requests.Response: The response as returned by python-requests.
        """
        auth = kwargs.pop("auth", False)
        if auth:
            self.ensure_auth()
            request_func = self.oidc.request
        else:
            request_func = requests.request
        try:
            response = request_func(verb, f"{self.base_url}{url}", **kwargs)
        except (OIDCClientError, ConnectionError) as e:
            raise BodhiClientException(str(e))
        if not response.ok:
            raise BodhiClientException(response.text)
        return munchify(response.json())

    def ensure_auth(self):
        """Make sure we are authenticated."""
        self.oidc.ensure_auth(use_kerberos=True)
        if not self.oidc.has_cookie("auth_tkt", domain=urlparse(self.base_url).hostname):
            while True:
                resp = self.oidc.request("GET", f"{self.base_url}oidc/login-token")
                if resp.ok:
                    break
                if resp.status_code == 401:
                    self.clear_auth()
                    self.oidc.login(use_kerberos=True)
                else:
                    resp.raise_for_status()

    def clear_auth(self):
        """Clear the authentication information."""
        self.oidc.clear_auth()
        self.csrf_token = ""

    @property
    def username(self):
        """Return the authenticated username."""
        return self.oidc.username

    @errorhandled
    def save(self, **kwargs) -> 'munch.Munch':
        """
        Save an update.

        This entails either creating a new update, or editing an existing one.
        To edit an existing update, you must specify the update alias in
        the ``edited`` keyword argument.

        Args:
            display_name (str): The name of the update.
            builds (str): A list of koji builds for this update.
            type (str): The type of this update: ``security``, ``bugfix``,
                ``enhancement``, and ``newpackage``.
            bugs (str): A list of Red Hat Bugzilla ID's associated with this
                update.
            notes (str): Details as to why this update exists.
            request (str): Request for this update to change state, either to
                ``testing``, ``stable``, ``unpush``, ``obsolete`` or None.
            close_bugs (bool): Close bugs when update is stable.
            suggest (str): Suggest that the user reboot or logout after update.
                (``reboot``, ``logout``).
            autotime (bool): Allow bodhi to automatically change the state of this
                update based on the time spent in testing by this update. It
                will push your update to ``stable`` once it reaches the ``stable_days``.
            stable_days (int): The minimun amount of time an update has to spend in
                ``testing`` before being automatically pushed to ``stable``.
            autokarma (bool): Allow bodhi to automatically change the state of this
                update based on the ``karma`` from user feedback.  It will
                push your update to ``stable`` once it reaches the ``stable_karma``
                and unpush your update when reaching ``unstable_karma``.
            stable_karma (int): The upper threshold for marking an update as
                ``stable``.
            unstable_karma (int): The lower threshold for unpushing an update.
            edited (str): The update alias of the existing update that we are
                editing.
            severity (str): The severity of this update (``urgent``, ``high``,
                ``medium``, ``low``).
            requirements (str): A list of required Taskotron tests that must pass
                for this update to reach stable. (e.g. ``dist.rpmdeplint``,
                ``dist.upgradepath``, ``dist.rpmlint``, etc).
            require_bugs (bool): A boolean to require that all of the bugs in your
                update have been confirmed by testers.
            require_testcases (bool): A boolean to require that this update passes
                all test cases before reaching stable.
            from_tag (str): The name of a Koji tag from which to pull builds
                instead of providing them manually in `builds`.
        Returns:
            The Bodhi server's response to the request.
        """
        kwargs['csrf_token'] = self.csrf()
        if 'type_' in kwargs:
            # backwards compat
            kwargs['type'] = kwargs['type_']
        return self.send_request('updates/', verb='POST', auth=True,
                                 data=kwargs)

    @errorhandled
    def request(self, update: str, request: str) -> 'munch.Munch':
        """
        Request an update state change.

        Args:
            update: The alias of the update.
            request: The request (``testing``, ``stable``, ``obsolete``, ``unpush``, ``revoke``).
        Returns:
            The response from Bodhi to the request.
        Raises:
            UpdateNotFound: If the server returns a 404 error code.
        """
        try:
            return self.send_request(f'updates/{update}/request',
                                     verb='POST', auth=True,
                                     data={'update': update, 'request': request,
                                           'csrf_token': self.csrf()})
        except RequestException as exc:
            if exc.response is not None and exc.response.status_code == 404:
                # The Bodhi server gave us a 404 on the resource, so let's raise an UpdateNotFound.
                raise UpdateNotFound(update)
            else:
                raise

    @errorhandled
    def waive(self, update: str, comment: str,
              tests: typing.Optional[typing.Iterable[str]] = None) -> 'munch.Munch':
        """
        Waive unsatisfied requirements on an update.

        Args:
            update: The alias of the update.
            comment: A comment explaining the waiver.
            tests: The list of unsatisfied requirements to waive. If not specified, all unsatisfied
                   requirements of this update will be waived.
        Returns:
            The response from Bodhi to the request.
        Raises:
            UpdateNotFound: If the server returns a 404 error code.
        """
        data = {'update': update, 'tests': tests, 'comment': comment, 'csrf_token': self.csrf()}
        try:
            return self.send_request(f'updates/{update}/waive-test-results',
                                     verb='POST', auth=True, data=data)
        except RequestException as exc:
            if exc.response is not None and exc.response.status_code == 404:
                # The Bodhi server gave us a 404 on the resource, so let's raise an UpdateNotFound.
                raise UpdateNotFound(update)
            else:
                raise

    @errorhandled
    def trigger_tests(self, update: str) -> 'munch.Munch':
        """
        Trigger tests for update.

        Args:
            update: The alias of the update to run tests for.
        Returns:
            The response from the post to trigger_tests/.
        """
        try:
            return self.send_request(
                f'updates/{update}/trigger-tests', verb='POST', auth=True,
                data={'update': update, 'csrf_token': self.csrf()})
        except RequestException as exc:
            if exc.response is not None and exc.response.status_code == 404:
                # The Bodhi server gave us a 404 on the resource, so let's raise an UpdateNotFound.
                raise UpdateNotFound(update)
            else:
                raise

    @errorhandled
    def query(self, **kwargs) -> 'munch.Munch':
        """
        Query bodhi for a list of updates.

        Args:
            alias (str): The update alias.
            updateid (str): The update ID (eg: FEDORA-2015-0001).
            content_type (str): A content type (rpm, module) to limit the query to.
            releases (str): A comma separated list of releases that you wish to query updates
                for.
            status (str): The update status (``pending``, ``testing``, ``stable``,
                ``obsolete``, ``unpushed``)
            type (str): The type of the update: ``security``, ``bugfix``,
                ``enhancement``, and ``newpackage``.
            bugs (str): A comma separated list of Red Hat Bugzilla IDs.
            request (str): An update request to query for
                ``testing``, ``stable``, ``unpush``, ``obsolete`` or None.
            mine (bool): If True, only query the users updates. Default: False.
            packages (str): A space or comma delimited list of package names.
            limit (int): A deprecated argument, sets ``rows_per_page``. See its docstring for more
                info.
            approved_before (str): A datetime string.
            approved_since (str): A datetime string.
            builds (str): A space or comma delimited string of build nvrs.
            critpath (bool): A boolean to query only critical path updates.
            locked (bool): A boolean to filter only locked updates.
            modified_before (str): A datetime string to query updates that have been modified
                before a certain time.
            modified_since (str): A datetime string to query updates that have been modified
                since a certain time.
            pushed (bool): A boolean to filter only pushed updates.
            pushed_before (str): A datetime string to filter updates pushed before a certain
                time.
            pushed_since (str): A datetime string to filter updates pushed since a
                certain time.
            severity (str): A severity type to filter by (``unspecified``,
                ``urgent``, ``high``, ``medium``, ``low``).
            submitted_before (str): A datetime string to filter updates submitted
                before a certain time.
            submitted_since (str): A datetime string to filter updates submitted
                after a certain time.
            suggest (str): Query for updates that suggest a user restart
                (``logout``, ``reboot``).
            gating (str): filter by TestGatingStatus description.
            from_side_tag (bool): A boolean to filter updates created from side tag or from
                normal workflow.
            user (str): Query for updates submitted by a specific user.
            rows_per_page (int): Limit the results to a certain number of rows per page
                (min:1 max: 100 default: 20).
            page (int): Return a specific page of results.
        Returns:
            The response from Bodhi describing the query results.
        """
        # bodhi1 compat
        if 'limit' in kwargs:
            kwargs['rows_per_page'] = kwargs['limit']
            del kwargs['limit']
        # 'mine' may be in kwargs, but set False
        if kwargs.get('mine'):
            if self.username is None:
                raise BodhiClientException("Could not get user info.")
            kwargs['user'] = self.username
        if 'package' in kwargs:
            # for Bodhi 1, 'package' could be a package name, build, or
            # update ID, so try and figure it out
            if re.search(UPDATE_TITLE_RE, kwargs['package']):
                kwargs['builds'] = kwargs['package']
            elif re.search(UPDATE_ID_RE, kwargs['package']):
                kwargs['updateid'] = kwargs['package']
            else:
                kwargs['packages'] = kwargs['package']
            del kwargs['package']
        if 'release' in kwargs:
            if isinstance(kwargs['release'], list):
                kwargs['releases'] = kwargs['release']
            else:
                kwargs['releases'] = [kwargs['release']]
            del kwargs['release']
        if 'type_' in kwargs:
            kwargs['type'] = kwargs['type_']
            del kwargs['type_']
        # Old Bodhi CLI set bugs default to "", but new Bodhi API
        # checks for 'if bugs is not None', not 'if not bugs'
        if 'bugs' in kwargs and kwargs['bugs'] == '':
            kwargs['bugs'] = None
        return self.send_request('updates/', verb='GET', params=kwargs)

    def get_test_status(self, update: str) -> 'munch.Munch':
        """
        Query bodhi for the test status of the specified update..

        Args:
            update: The alias of the update to retrieve the test status of.
        Returns:
            The response from Bodhi describing the query results.
        """
        return self.send_request(f'updates/{update}/get-test-results', verb='GET')

    @errorhandled
    def comment(self, update: str, comment: str, karma: int = 0) -> 'munch.Munch':
        """
        Add a comment to an update.

        Args:
            update: The alias of the update comment on.
            comment: The text of the comment to add to the update.
            karma: The amount of karma to leave. May be -1, 0, or 1. Defaults to 0.
        Returns:
            The response from the post to comments/.
        """
        return self.send_request(
            'comments/', verb='POST', auth=True,
            data={'update': update, 'text': comment, 'karma': karma, 'csrf_token': self.csrf()})

    @errorhandled
    def save_override(
        self, nvr: str,
        notes: str,
        duration: typing.Optional[int] = None,
        expiration_date: typing.Optional[datetime.datetime] = None,
        edit: bool = False,
        expired: bool = False
    ) -> 'munch.Munch':
        """
        Save a buildroot override.

        This entails either creating a new buildroot override, or editing an
        existing one.

        Args:
            nvr: The nvr of a koji build.
            notes: Notes about why this override is in place.
            duration: Number of days from now that this override should expire.
            expiration_date: Date when this override should expire. This argument or the
                ``duration`` argument must be provided.
            edit: True if we are editing an existing override, False otherwise. Defaults to False.
            expired: Set to True to expire an override. Defaults to False.
        Returns:
            A dictionary-like representation of the saved override.
        """
        if duration is None and expiration_date is None:
            raise TypeError("The duration or the expiration_date must be provided.")
        if duration is not None and expiration_date is not None:
            raise TypeError(
                "The duration and the expiration_date cannot be provided at the same time."
            )
        if duration:
            expiration_date = datetime.datetime.utcnow() + \
                datetime.timedelta(days=duration)
        data = {'nvr': nvr,
                'expiration_date': expiration_date,
                'notes': notes,
                'csrf_token': self.csrf()}
        if edit:
            data['edited'] = nvr
        if expired:
            data['expired'] = expired
        return self.send_request(
            'overrides/', verb='POST', auth=True, data=data)

    @errorhandled
    def get_compose(self, release: str, request: str) -> 'munch.Munch':
        """
        Get information about compose.

        Args:
            release: The name of the release.
            request: The request (``testing``, ``stable``).
        Returns:
            The response from Bodhi to the request.
        Raises:
            ComposeNotFound: If the server returns a 404 error code.
        """
        try:
            return self.send_request(f'composes/{release}/{request}', verb='GET')
        except RequestException as exc:
            if exc.response is not None and exc.response.status_code == 404:
                # The Bodhi server gave us a 404 on the resource, so let's raise an ComposeNotFound.
                raise ComposeNotFound(release, request)
            else:
                raise

    @errorhandled
    def list_composes(self) -> 'munch.Munch':
        """
        List composes.

        Returns:
            A dictionary-like representation of the Composes.
        """
        return self.send_request('composes/', verb='GET')

    @errorhandled
    def list_overrides(
            self, user: typing.Optional[str] = None, packages: typing.Optional[str] = None,
            expired: typing.Optional[bool] = None, releases: typing.Optional[str] = None,
            builds: typing.Optional[str] = None, rows_per_page: typing.Optional[int] = None,
            page: typing.Optional[int] = None) -> 'munch.Munch':
        """
        List buildroot overrides.

        Args:
            user: A username whose buildroot overrides you want returned.
            packages: Comma separated package names to filter buildroot overrides by.
            expired: If True, only return expired overrides. If False, only return active
                overrides.
            releases: Comma separated Release shortnames to filter buildroot overrides by.
            builds: Comma separated build NVRs to filter overrides by.
            rows_per_page: Limit the results to a certain number of rows per page.
            page: Return a specific page of results.
        Returns:
            A dictionary-like representation of the Overrides.
        """
        params: typing.MutableMapping[str, typing.Union[int, str, None]] = {}
        if user:
            params['user'] = user
        if packages:
            params['packages'] = packages
        if expired is not None:
            params['expired'] = expired
        if releases:
            params['releases'] = releases
        if builds:
            params['builds'] = builds
        if rows_per_page:
            params['rows_per_page'] = rows_per_page
        if page:
            params['page'] = page
        return self.send_request('overrides/', verb='GET', params=params)

    @errorhandled
    def csrf(self) -> str:
        """
        Return CSRF token if already acquired, otherwise login, get a CSRF, cache it, and return.

        If there is already a CSRF token, this method returns it.

        If there is not, this method ensures that we know the username, logs in if we aren't already
        logged in acquires and caches a CSRF token, and returns it.

        Returns:
            The CSRF token.
        """
        if not self.csrf_token:
            self.csrf_token = self.send_request(
                'csrf', verb='GET', auth=True)['csrf_token']
        return self.csrf_token

    def parse_file(self, input_file: str) -> typing.List[typing.Dict[str, typing.Any]]:
        """
        Parse an update template file.

        Args:
            input_file: The filename of the update template.
        Returns:
            A list of dictionaries of parsed update values which
                can be directly passed to the ``save`` method.
        Raises:
            ValueError: If the ``input_file`` does not exist, or if it cannot be parsed.
        """
        if not os.path.exists(input_file):
            raise ValueError(f"No such file or directory: {input_file}")

        defaults = dict(severity='unspecified', suggest='unspecified')
        config = configparser.ConfigParser(defaults=defaults)
        read = config.read(input_file)

        if len(read) != 1 or read[0] != input_file:
            raise ValueError(f"Invalid input file: {input_file}")

        updates = []

        for section in config.sections():
            update = {
                'builds': section,
                'bugs': config.get(section, 'bugs', raw=True),
                'close_bugs': config.getboolean(section, 'close_bugs'),
                'display_name': config.get(section, 'display_name', raw=True, fallback=None),
                'type': config.get(section, 'type', raw=True),
                'type_': config.get(section, 'type', raw=True),
                'request': config.get(section, 'request', raw=True),
                'severity': config.get(section, 'severity', raw=True),
                'notes': config.get(section, 'notes', raw=True),
                'autokarma': config.get(section, 'autokarma', raw=True),
                'stable_karma': config.get(section, 'stable_karma', raw=True),
                'unstable_karma': config.get(
                    section, 'unstable_karma', raw=True),
                'suggest': config.get(section, 'suggest', raw=True)}

            updates.append(update)

        return updates

    @errorhandled
    def latest_builds(self, package: str) -> 'munch.Munch':
        """
        Get the latest builds for a package.

        Args:
            package: The package name, for example "kernel".
        Returns:
            A dict-like object of the release dist tag to the latest build.
        """
        return self.send_request('latest_builds', params={'package': package})

    def testable(self) -> typing.Iterator[dict]:
        """
        Return a generator that iterates installed testing updates.

        This method is a generator that yields packages that you currently
        have installed that you have yet to test and provide feedback for.

        Only works on systems with dnf.

        Returns:
            generator: An iterable of dictionaries describing updates that match builds installed on
                the local system.
        Raises:
            RuntimeError: If the dnf Python bindings are not installed.
        """
        if dnf is None:
            raise RuntimeError('dnf is required by this method and is not installed.')

        base = dnf.Base()
        sack = base.fill_sack(load_system_repo=True)
        query = sack.query()
        installed = query.installed()
        with open('/etc/fedora-release', 'r') as f:
            fedora = f.readlines()[0].split()[2]
        tag = f'f{fedora}-updates-testing'
        builds = self.get_koji_session().listTagged(tag, latest=True)
        for build in builds:
            pkgs = installed.filter(name=build['name'], version=build['version'],
                                    release=build['release']).run()
            if len(pkgs):
                update_list = self.query(builds=build['nvr'])['updates']
                for update in update_list:
                    yield update

    @staticmethod
    def compose_str(compose: dict, minimal: bool = True) -> str:
        """
        Return a string representation of a compose.

        Args:
            compose: A dictionary representation of a Compose.
            minimal: If True, return a minimal one-line representation of the compose.
                Otherwise, return a more verbose string. Defaults to True.
        Returns:
            A human readable string describing the compose.
        """
        line_formatter = '{0:<16}: {1}'
        security = '*' if compose['security'] else ' '
        title = f"{security}{compose['release']['name']}-{compose['request']}"
        details = f"{len(compose['update_summary']):3d} updates ({compose['state']}) "
        minimal_repr = line_formatter.format(title, details)

        if minimal:
            return minimal_repr

        line_formatter = '{0:>12}: {1}\n'

        compose_lines = [f'{"=":=^80}\n', f'     {minimal_repr}\n', f'{"=":=^80}\n']

        compose_lines += [
            line_formatter.format('Content Type', compose['content_type']),
            line_formatter.format('Started', compose['date_created']),
            line_formatter.format('Updated', compose['state_date']),
        ]

        if 'error_message' in compose and compose['error_message']:
            compose_lines.append(line_formatter.format('Error', compose['error_message']))

        compose_lines += ['\nUpdates:\n\n']
        line_formatter = f'\t{line_formatter}'
        for s in compose['update_summary']:
            compose_lines.append(line_formatter.format(s['alias'], s['title']))

        return ''.join(compose_lines)

    @staticmethod
    def override_str(override: dict, minimal: bool = True) -> str:
        """
        Return a string representation of a given override dictionary.

        Args:
            override: An override dictionary.
            minimal: If True, return a minimal one-line representation of the override.
                Otherwise, return a more verbose string. Defaults to True.
        Returns:
            A human readable string describing the given override.
        """
        if isinstance(override, str):
            return override

        if minimal:
            return (f"{override['submitter']['name']}'s {override['build']['nvr']} override "
                    f"(expires {override['expiration_date']})")

        divider = '=' * 60
        nvr = '\n'.join(textwrap.wrap(override['build']['nvr'].replace(',', ', '), width=60,
                        initial_indent=' ' * 5, subsequent_indent=' ' * 5))
        val = f"{divider}\n{nvr}\n{divider}\n"
        val += f"  Submitter: {override['submitter']['name']}\n"
        val += f"  Expiration Date: {override['expiration_date']}\n"
        val += f"  Notes: {override['notes']}\n"
        val += f"  Expired: {override['expired_date'] is not None}"

        return val

    def update_str(self, update: dict, minimal: bool = False) -> str:
        """
        Return a string representation of a given update dictionary.

        Args:
            update: An update dictionary, acquired by the ``list`` method.
            minimal: If True, return a minimal one-line representation of the update.
                Otherwise, return a more verbose representation. Defaults to False.
        Returns:
            A human readable string describing the given update.
        """
        if isinstance(update, str):
            return update
        if minimal:
            val = ""
            security = '*' if update['type'] == 'security' else ' '
            date = update['date_pushed'] and update['date_pushed'].split()[0] \
                or update['date_submitted'].split()[0]
            days_in_status = _days_since(update['date_pushed']) if update['date_pushed'] \
                else _days_since(update['date_submitted'])
            if update['builds']:
                title = update['builds'][0]['nvr']
            else:
                title = update['title'] or update['alias']
            content_type = update['content_type'] or 'unspecified'
            val += (f"{security}{title:40} {content_type:9}  "
                    f"{update['status']:8}  {date:>10} ({days_in_status})")
            for build in update['builds'][1:]:
                val += f"\n  {build['nvr']}"
            return val

        # Content will be formatted as wrapped lines, each line is in format
        #    indent            content wrap width
        #  |--> 12 <--|  |-->        66       .... <--|
        # "            : wrapped line ...              "
        #  |-->          80 chars in total         <--|
        wrap_width = 66
        wrap_line = functools.partial(textwrap.wrap, width=wrap_width)
        line_formatter = '{0:>12}: {1}\n'

        update_lines = [f'{"=":=^80}\n']
        update_lines += [
            line + '\n' for line in textwrap.wrap(
                update['title'],
                width=80,
                initial_indent=' ' * 5,
                subsequent_indent=' ' * 5)
        ]
        update_lines.append(f'{"=":=^80}\n')

        update_lines.append(
            line_formatter.format('Update ID', update['alias']))

        update_lines += [
            line_formatter.format('Content Type', update['content_type']),
            line_formatter.format('Release', update['release']['long_name']),
            line_formatter.format('Status', update['status']),
            line_formatter.format('Type', update['type']),
            line_formatter.format('Severity', update['severity']),
            line_formatter.format('Karma', update['karma']),
            line_formatter.format(
                'Autokarma',
                f"{update['autokarma']}  [{update['unstable_karma']}, {update['stable_karma']}]"),
            line_formatter.format('Autotime', update['autotime'])
        ]

        try:
            test_status = self.get_test_status(update['alias'])
        except RequestException as err:
            log.debug('ERROR while retrieving CI status: %s', err)
            test_status = None

        if test_status:
            info = None
            waivers = None
            if 'errors' in test_status:
                info = '\n'.join([el.description for el in test_status.errors])
            elif 'decision' in test_status:
                info = test_status.decision.summary
                waivers = test_status.decision.waivers
            elif 'decisions' in test_status:
                info = '\n'.join([d.summary for d in test_status.decisions])
                waivers = list(itertools.chain(*(d.waivers for d in test_status.decisions)))
            else:
                log.debug('No `errors` nor `decision` in the data returned')
            if info:
                update_lines.append(line_formatter.format('CI Status', info))
            if waivers:
                waivers_lines = []
                for waiver in waivers:
                    dt = datetime.datetime.strptime(waiver['timestamp'], '%Y-%m-%dT%H:%M:%S.%f')
                    waivers_lines.append(
                        f"{waiver['username']} - {dt.strftime('%Y-%m-%d %H:%M:%S')}")
                    waivers_lines += wrap_line(waiver['comment'])
                    waivers_lines.append(f"build: {waiver['subject_identifier']}")
                    waivers_lines.append(f"testcase: {waiver['testcase']}")

                update_lines.append(line_formatter.format('Waivers', waivers_lines[0]))
                waiver_line_formatter = line_formatter.replace(': ', '  ')
                update_lines += [
                    waiver_line_formatter.format(indent, line)
                    for indent, line in zip(
                        itertools.repeat(' ', len(waivers_lines) - 1),
                        waivers_lines[1:])
                ]

        if update['request'] is not None:
            update_lines.append(line_formatter.format('Request', update['request']))

        if len(update['bugs']):
            bugs = list(itertools.chain(*[
                wrap_line(f"{bug['bug_id']} - {bug['title']}")
                for bug in update['bugs']
            ]))
            indent_lines = ['Bugs'] + [' '] * (len(bugs) - 1)
            update_lines += [
                line_formatter.format(indent, line)
                for indent, line in zip(indent_lines, bugs)
            ]

        if update['notes']:
            notes_lines = list(itertools.chain(
                *[wrap_line(line) for line in update['notes'].splitlines()]
            ))
            indent_lines = ['Notes'] + [' '] * (len(notes_lines) - 1)
            for indent, line in zip(indent_lines, notes_lines):
                update_lines.append(line_formatter.format(indent, line))

        update_lines += [
            line_formatter.format('Submitter', update['user']['name']),
            line_formatter.format('Submitted', update['date_submitted']),
        ]

        if len(update['comments']):
            comments_lines = []
            for comment in update['comments']:
                comments_lines.append(
                    f"{comment['user']['name']} - {comment['timestamp']} "
                    f"(karma {comment['karma']})")
                comments_lines += wrap_line(comment['text'])

            update_lines.append(line_formatter.format('Comments', comments_lines[0]))
            comment_line_formatter = line_formatter.replace(': ', '  ')
            update_lines += [
                comment_line_formatter.format(indent, line)
                for indent, line in zip(
                    itertools.repeat(' ', len(comments_lines) - 1),
                    comments_lines[1:])
            ]

        update_lines.append(
            f"\n  {self.base_url}updates/{update['alias']}\n")

        return ''.join(update_lines)

    @errorhandled
    def get_releases(self, **kwargs) -> 'munch.Munch':
        """
        Return a list of bodhi releases.

        This method returns a dictionary in the following format::

            {"releases": [
                {"dist_tag": "dist-f12", "id_prefix": "FEDORA",
                 "locked": false, "name": "F12", "long_name": "Fedora 12"}]}

        Args:
            kwargs: A dictionary of extra parameters to pass along with the request.
        Returns:
            A dictionary describing Bodhi's release objects.
        """
        return self.send_request('releases/', verb='GET', params=kwargs)

    def get_koji_session(self) -> koji.ClientSession:
        """
        Return an authenticated koji session.

        Returns:
            An initialized authenticated koji client.
        """
        config = configparser.ConfigParser()
        koji_conf = os.path.join(os.path.expanduser('~'), '.koji', 'config')
        if not os.path.exists(koji_conf):
            koji_conf = '/etc/koji.conf'
        with open(koji_conf) as fh:
            config.read_file(fh)
        session = koji.ClientSession(config.get('koji', 'server'))
        return session

    koji_session = property(fget=get_koji_session)

    def candidates(self) -> typing.Iterable[dict]:
        """
        Get a list list of update candidates.

        Returns:
            A list of koji builds (dictionaries returned by koji.listTagged()) that are tagged
            as candidate builds and are owned by the current user.
        """
        builds = []
        data = self.get_releases()
        koji = self.get_koji_session()
        for release in data['releases']:
            try:
                for build in koji.listTagged(release['candidate_tag'], latest=True):
                    if build['owner_name'] == self.username:
                        builds.append(build)
            except Exception:
                log.exception('Unable to query candidate builds for %s', release)
        return builds
